/* ScummVM - Scumm Interpreter
 * PS2 port
 * Copyright (C) 2002  Marcus Comstedt
 * For PS2 altered by Ole
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header: /cvsroot/scummvm/scummvm/backends/dc/audio.cpp,v 1.6 2003/06/18 21:15:08 marcus_c Exp $
 *
 */

#include <common/stdafx.h>
#include <common/scummsys.h>
#include <base/engine.h>
#include <loadfile.h>
#include "ps2.h"
#include "g2lib/g2.h"
#include "sjpcm.h"
#include "audio.h"


static int frame_size = 800; //NTSC = 800, PAL = 960

void swapBytes(short* buff, int size);

void play_sound() {
   extern bool* gPlaybackStopped;

   if (!*gPlaybackStopped) return;
   SjPCM_Play(); 
   *gPlaybackStopped = false;
}

void pause_sound() {
   extern bool* gPlaybackStopped;

   if (*gPlaybackStopped) return;
   SjPCM_Pause();
   *gPlaybackStopped = true;

}

bool OSystem_Ps2::setSoundCallback(SoundProc proc, void *param)
{
  _sound_proc_param = param;
  _sound_proc = proc;
  _ps2_sound_buffer = 1;
  return true;
}

void OSystem_Ps2::clearSoundCallback()
{
  _sound_proc = NULL;
  _sound_proc_param = NULL;
}

int OSystem_Ps2::getOutputSampleRate() const
{
  return 24000;
}


void OSystem_Ps2::playSound()
{
   play_sound();
}

void OSystem_Ps2::pauseSound()
{
   pause_sound(); 
}


/*

Note: we direct the scummvm to generate 24000Hz samples even though the
sjpcm needs 48000Hz samples. The trick/hack is caused by the fact that
sjpcm wants 2 separate sound buffers for left and right channels. But
scummvm genrates the sound data into single buffer, so some sort of 
buffer splitting would be needed. Instead of splitting we left the STEREO
sound data as they are (generated at 24000Hz)  and feed the sjpcm sound lib
by the same buffer twice. The same buffer for left and right channel.
That way we gain:
- speed by generating less sound data (note addlib is emulated)
- memory by not using large sound buffers
- speed by not transferring large emounts of data
and we lost:
- the STEREO effect (we send the same buffer to both L & R channels)
- small amount of sound precision (L&R might sometimes hold different 
  data, wchich we mix into single channel)
*/

void OSystem_Ps2::checkSound()
{
#ifndef PCSX2


  int buffered;
  int i;
  int sendFrames = 5; 
  int treshold;
  short* play_buff = temp_sound_buffer[_active_buffer];
  short* store_buff = temp_sound_buffer[1-_active_buffer];

  if (_playback_stopped) return;

  buffered = SjPCM_Buffered();

  treshold =  _ps2_sound_buffer* _hires;
  if (treshold > 6) treshold = 6;
  treshold *=frame_size;

  // the hardware buffer is nearly empty (less than 2 frames)
  // feed him the next data from play buffer
  // _ps2_sound buffer is generally number of frames. For common
  // scenes it may be small (1 or 2) but for smush playbacks the 
  // value should be about 5. The value is switched by the
  // setFeatureState method.
  // _hires is either 1(lowres) or 2 (hires)
  // hires takes more time to render so we double the threshold
  if (buffered < treshold) {

   //note: we put stereo buffer in the left and right channel!
    SjPCM_Enqueue(play_buff,play_buff, frame_size*sendFrames, 0); 
       store_buff = temp_sound_buffer[_active_buffer];
       _active_buffer = 1-_active_buffer;

	//we get stereo sample encoded in one buffer
       _sound_proc(_sound_proc_param, (byte*)store_buff , frame_size*2*sendFrames);
  }
  
#endif
}

void ps2_init_sound(const char* path1, const char* path2, int mode) {
#ifndef PCSX2
    int ret;
    char moduleName[64];

    if (mode == 0 ) {
      frame_size = 800; //NTSC
    } else {
      frame_size = 960; //PAL
    }
  

    if (path1 != NULL) {
	    strcpy(moduleName, path1);
	    strcat(moduleName, "SJPCM.IRX");
	    if (path1[0] == 'c' && path1[2] == 'r') {
		strcat(moduleName, ";1");
	    }

	    //SJPCM requires rom0:LIBSD - it is loaded elsewhere
	    ret = SifLoadModule(moduleName,0, NULL);
	    if (ret < 0) {
		strcpy(moduleName, path2);
		strcat(moduleName, "SJPCM.IRX");
   		if (path2[0] == 'c' && path2[2] == 'r') {
			strcat(moduleName, ";1");
		}
		ret = SifLoadModule(moduleName,0, NULL);
		if (ret < 0) {
			printf("sifLoadModule %s failed: %d\n", moduleName, ret);
			while(1);
		}
	    }

	    //short delay - give the module enough time to register itself
	    ret = 0x01000000;
	    while(ret--) asm("nop\nnop\nnop\nnop");

	    //non-sync mode
	    SjPCM_Init(0, mode);
	    SjPCM_Setvol(0x3fff);
	}
#endif
}
