/* ScummVM - Scumm Interpreter
 * Dreamcast port
 * Copyright (C) 2002  Marcus Comstedt
 * For PS2 altered by Ole
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header: /cvsroot/scummvm/scummvm/backends/dc/display.cpp,v 1.10 2003/07/14 21:44:01 marcus_c Exp $
 *
 */
#include <common/stdafx.h>
#include <common/scummsys.h>
#include <base/engine.h>
#include "ps2.h"
#include "g2lib/g2.h"


void OSystem_Ps2::setVideoResolution(uint w, uint h) {
	int pal_mode;
        int hires;

	// clear the screen - screen must have been initialized already
	g2_set_fill_color(0, 0, 0);
	g2_fill_rect(0, 0, g2_get_max_x(), g2_get_max_y());

	pal_mode = g2_is_pal();
	hires = 0;
	if (w > 320 && h > 256) {
		hires = 1;
	}
	_hires = hires  + 1;
	//_hires = 1;

	if (pal_mode) {
		if (hires) {
			g2_init(PAL_640_512_32);
			_top_screen = (512 - h)/2;
			//g2_init(PAL_640_256_32);
		} else {
			g2_init(PAL_320_256_32);
			_top_screen = (256 - h)/2;
		}
	} else {
		if (hires){
	        	g2_init(NTSC_640_448_32);
			_top_screen = 12;
		} else {
			g2_init(NTSC_320_224_32);
			_top_screen = (240 - h)/2;
		}
	}

	// clear the screen
	g2_set_fill_color(0, 0, 0);
	g2_fill_rect(0, 0, g2_get_max_x(), g2_get_max_y());
}


void OSystem_Ps2::setPalette(const byte *colors, uint start, uint num)
{
  uint32* dst = palette + start;
  uint32 r,g,b;

  if(num>0)
    while( num-- ) {
      r = colors[0];
      g = colors[1];
      b = colors[2]; 
      *dst = 	r | g<<8 | b<<16;
      colors += 4;
      dst++;
    }
_screen_dirty = true;
}

/*
void OSystem_Ps2::setScaling() {
	//no scaling at the moment
}
*/

void OSystem_Ps2::initSize(uint w, uint h)
{
  int i;
  assert(w <= SCREEN_W && h <= SCREEN_H);

//  printf("display: start init size: w=%i h=%i\n", w, h);

  setVideoResolution(w,h);

  _overlay_visible = false;
  _screen_w = w;
  _screen_h = h;
  _overlay_x = (w-OVL_W)/2;
  _overlay_y = (h-OVL_H)/2;
  if(_overlay_x<0) _overlay_x = 0;
  if(_overlay_y<0) _overlay_y = 0;

  if(!screen) {
    //screen = new unsigned char[SCREEN_W*SCREEN_H];
	screen = new unsigned char[_screen_w*_screen_h];
	memset(screen, 0, _screen_w*_screen_h);
  }

//  printf("display: test  mouse_bg=%p mouse_fg=%p \n", backg_tx, mouse_tx);

  if(!mouse_tx)
    mouse_tx = new uint32[MOUSE_W*MOUSE_H];

  if(!backg_tx)
    backg_tx = new uint32[MOUSE_W*MOUSE_H];

  if(!overlay) {
    overlay = new unsigned short[_screen_w*_screen_h/*OVL_W*OVL_H*/];
    memset(overlay, 0, _screen_w*_screen_h*2);
  }


  _screen_buffer = 0;
  _mouse_buffer = 0;
  _overlay_buffer = 0;
  _screen_dirty = true;
  _overlay_dirty = true;

  _ms_cur_x = w / 2;
  _ms_cur_y = h / 2;

  for (i = 0; i < MOUSE_W*MOUSE_H; i++) {
    mouse_tx[i] = 10;
    backg_tx[i] = 10;
  }
  
}

void OSystem_Ps2::copyRectToScreen(const byte *buf, int pitch, int x, int y,
				  int w, int h)
{

  /* Clip the coordinates */
  if (x < 0) { w+=x; buf-=x; x = 0; }
  if (y < 0) { h+=y; buf-=y*pitch; y = 0; }
  if (w > _screen_w-x) { w = _screen_w - x; }
  if (h > _screen_h-y) { h = _screen_h - y; }
			
  if (w <= 0 || h <= 0)
	return;

  unsigned char *dst = screen + y*_screen_w + x;
  do {
    memcpy(dst, buf, w);
    dst += _screen_w;
    buf += pitch;
  } while (--h);
  _screen_dirty = true;
}


bool OSystem_Ps2::showMouse(bool visible)
{	
  bool last = _ms_visible;
  _ms_visible = visible;

  return last;
}

void OSystem_Ps2::warpMouse(int x, int y)
{

  if (_overlay_visible) {
	drawMouse(_ms_cur_x, _ms_cur_y, _ms_cur_w, _ms_cur_h, _ms_buf, _ms_visible);   
  } 
  _ms_cur_x = x;
  _ms_cur_y = y;

  if (_ms_cur_x > _screen_w) {
	_ms_cur_x = _screen_w;
  }
  if (_ms_cur_y > _screen_h) {
	_ms_cur_y = _screen_h;
  }

}

void OSystem_Ps2::setMouseCursor(const byte *buf, uint w, uint h,
					 int hotspot_x, int hotspot_y, byte keycolor)
{
  _ms_cur_w = w;
  _ms_cur_h = h;
  if (_ms_cur_x > _screen_w) {
	_ms_cur_x = _screen_w;
  }
  if (_ms_cur_y > _screen_h) {
	_ms_cur_y = _screen_h;
  }
  
  _ms_hotspot_x = hotspot_x;
  _ms_hotspot_y = hotspot_y;
  
  _ms_buf = (byte*)buf;
  _ms_keycolor = keycolor;
//  printf("Set mouse cursor hotspot x=%i y=%i \n", hotspot_x, hotspot_y);

}

void OSystem_Ps2::setShakePos(int shake_pos)
{
  _current_shake_pos = shake_pos;
}

void OSystem_Ps2::updateScreen(void)
{
  int top;
//  int time;

  if (	_hires == 1) {
	g2_wait_vsync();
  }

  if(_screen_dirty && !_overlay_visible) {
	//clear the black space under the screen
	top = _top_screen + _current_shake_pos;  //top of the screen
	g2_fill_rect(0, _screen_h + top, _screen_w, _screen_h + top+16 ); 
	//display image
//	time = getMillis();
	g2_put_image8(0, top, _screen_w, _screen_h, screen, palette, 1/*_hires*/);
//	time = getMillis() - time;
//	printf("putImage time=%d \n", time);
	if (_current_shake_pos) {
		//clear the black space atop of screen
		g2_fill_rect(0,_top_screen, _screen_w, top -1); 
	}
	_screen_dirty = false;

	// Hack for sound playback :
	// overlay is not visible but the sound is stopped -> play again
	if ( _playback_stopped)  playSound();
  } else
  if( _overlay_dirty && _overlay_visible ) {
//	time = getMillis();
	g2_put_image16(0,_top_screen, _screen_w, _screen_h, overlay);
//	time = getMillis() - time;
//	printf("putOverlay time=%d \n", time);
	_overlay_dirty = false;
	if ( _playback_stopped)  playSound();
  }
//  printf("update screen  mouse x=%d y=%d visible=%d\n", _ms_cur_x, _ms_cur_y, _ms_visible);

  drawMouse(_ms_cur_x, _ms_cur_y, _ms_cur_w, _ms_cur_h, _ms_buf, _ms_visible);
}


void OSystem_Ps2::getMouseBackground(void) {
	uint32 *dst = backg_tx;
	int y;
	int x;
	unsigned short pix16;
	int screenOffs;


        //printf("lx=%d, ly=%d, lw=%d, lh=%d \n", _ms_last_x, _ms_last_y, _ms_last_w, _ms_last_h);
	if (!_overlay_visible) {    //get mouse background from regular screen
	    for(y=0; y<_ms_last_h; y++) {
	      for(x=0; x<_ms_last_w; x++) {
		  screenOffs = (y+ _ms_last_y) * _screen_w + (x +_ms_last_x);
		  *dst++ = palette[screen[screenOffs]];
              }
	    }

	} else {  //overlay visible - get mouse background from overlay
	    for(y=0; y<_ms_last_h; y++) {
	      for(x=0; x<_ms_last_w; x++) {
		//TODO - ???
		  screenOffs = (y+ _ms_last_y) * _screen_w + (x +_ms_last_x);
	          pix16 = overlay[screenOffs];
		  *dst++ = ((pix16&0x1f)<<3) | ((pix16&0x7E0)<<5) | ((pix16&0xF800)<<8);
              }
	    }
	}

}

void OSystem_Ps2::drawMouse(int xdraw, int ydraw, int w, int h,
				  unsigned char *buf, bool visible)
{

  uint32 *dst = mouse_tx;
  int y=0;
  int x;
  int screenOffs;
  unsigned short pix16;
  int mouseWidth; 
  int mouseHeight;
  int xshift = 0;
  int yshift = 0;


// Hack for sound playback :
// overlay is not visible but the sound is stopped -> play again
//  if (!_overlay_visible && _playback_stopped)  playSound();
    
//  printf("draw mouse x=%i y=%i w=%i h=%i  hotx=%i hoty=%i vis=%d buf=%p\n ", xdraw, ydraw, w,h, _ms_hotspot_x, _ms_hotspot_y, visible, buf);
 
//  if(visible && w<=MOUSE_W && h<=MOUSE_H) {
  if(visible && w*h <= MOUSE_W * MOUSE_H) {

        mouseWidth = w;
        mouseHeight= h;

        xshift = 0;
        yshift = 0;
	xdraw-= _ms_hotspot_x;
	ydraw-= _ms_hotspot_y;

        if (xdraw < 0) {
           xshift = -xdraw;
           mouseWidth-= xshift;   //decrease mouse width 
           xdraw = 0;
        }
        if (ydraw < 0) {
           yshift = -ydraw;
           mouseHeight-=  yshift; //decrease mouse height
           ydraw = 0;
        }


	//calculate clipping borders of the mouse according the position
        if (xdraw+w > _screen_w) {
            mouseWidth = _screen_w - xdraw;
        }
        if (ydraw+h > _screen_h) {
            //printf("mouse cursor bottom clip \n");
            mouseHeight = _screen_h - ydraw;
        }
        if (mouseWidth <=0 || mouseHeight<=0) {
            return; 
        }


        getMouseBackground();

        
        //if shifts are active - modify mouse buffer start
        buf+= (w* yshift)  + xshift;

	if (!_overlay_visible) {    //get mouse background from regular screen
	    for(y=0; y<mouseHeight; y++) {
	      for(x=0; x<mouseWidth; x++) 
		if(*buf == _ms_keycolor) {
		  screenOffs = (y+ ydraw) * _screen_w + (x + xdraw);
		  *dst++ = palette[screen[screenOffs]];
		  buf++;
		} else    *dst++ = palette[*buf++];
              buf+=(w-mouseWidth);	        
	    }

	} else {  //overlay visible - get mouse background from overlay
	    for(y=0; y<mouseHeight; y++) {
	      for(x=0; x<mouseWidth; x++) 
		if(*buf == 0xff) {
		//TODO - ???? - overlay and SCREEN_W
		  screenOffs = (y+ ydraw) * _screen_w + (x + xdraw);
                  pix16 = overlay[screenOffs];
		  *dst++ = ((pix16&0x1f)<<3) | ((pix16&0x7E8)<<5) | ((pix16&0xF800)<<8);
		  buf++;
		} else    *dst++ = palette[*buf++];
              buf+=(w-mouseWidth);	        
	    }
	}


      //g2_wait_vsync();
      g2_put_image(_ms_last_x, _ms_last_y+ _top_screen, _ms_last_w, _ms_last_h, backg_tx,/*_hires*/ 1);
      g2_put_image(xdraw, ydraw+_top_screen, mouseWidth, mouseHeight, mouse_tx, /*_hires */ 1);
	
      _ms_last_x = xdraw;
      _ms_last_y = ydraw;
      _ms_last_w = mouseWidth;
      _ms_last_h = mouseHeight;
  }
}


void OSystem_Ps2::showOverlay()
{
  _overlay_visible = true;
  clearOverlay();
}

void OSystem_Ps2::hideOverlay()
{
  _overlay_visible = false;
  _screen_dirty = true;
}

void OSystem_Ps2::clearOverlay()
{
  if(!_overlay_visible)
    return;

  unsigned char *src = screen;
  unsigned short *dst = overlay;
  uint32 pix32;
//  printf("display: clear_overlay...   ovl_x=%i  ovl_y=%i\n", _overlay_x, _overlay_y);

  for(int y=0; y<_screen_h; y++) {
    for(int x=0; x< _screen_w; x++) {
      pix32 = palette[src[x]];
      dst[x] = ((pix32 & 0xF8) >>3) | ((pix32 & 0xFC00) >> 5) | ((pix32 & 0xF80000) >>8);
    }
    src += _screen_w;
    dst += _screen_w;
  }
  _overlay_dirty = true;
}

void OSystem_Ps2::grabOverlay(int16 *buf, int pitch)
{
  int h = _screen_h /*OVL_H */;
  unsigned short *src = overlay;
  int width = _screen_w /*OVL_W */ *sizeof(int16);

  //printf("grabOverlay: pitch=%d\n", pitch);
  do {
    memcpy(buf, src, width);
    src += _screen_w/*OVL_W*/;
    buf += pitch;
  } while (--h);
}

void OSystem_Ps2::copyRectToOverlay(const int16 *buf, int pitch,
					  int x, int y, int w, int h)
{
  unsigned short *dst = overlay + y* _screen_w + x;
//  printf("copy rect to overlay: x=%d y=%d w=%d  h=%d pitch=%d \n", x,y, w,h, pitch);

  int width = w*sizeof(int16);
  do {
    memcpy(dst, buf, width);
    dst += _screen_w;
    buf += pitch;
  } while (--h);
  _overlay_dirty = true;
}

static const OSystem::GraphicsMode gfxmodes[] = {
  { "default", "320x200 32bpp", 0 },
  { NULL, NULL, 0 }
};

const OSystem::GraphicsMode *OSystem_Ps2::getSupportedGraphicsModes() const
{
  return gfxmodes;
}

int OSystem_Ps2::getDefaultGraphicsMode() const
{
  return 0;
}

bool OSystem_Ps2::setGraphicsMode(int mode)
{
  return mode == 0;
}

int OSystem_Ps2::getGraphicsMode() const
{
  return 0;
}
