/* ScummVM - Scumm Interpreter
 * PS2 port 
 * Copyright (C) 2002  Marcus Comstedt
 * For PS2 altered by Ole
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header: /cvsroot/scummvm/scummvm/backends/ps2/ps2main.cpp,v 1.14.2.2 2003/07/28 22:25:12 marcus_c Exp $
 *
 */


#define REAL_MAIN

#include <common/stdafx.h>
#include <common/scummsys.h>
#include <base/engine.h>
#include <base/gameDetector.h>
#include <base/plugins.h>
#include <kernel.h>

#include <tamtypes.h>
#include <sifcmd.h>
#include <kernel.h>
#include <sifrpc.h>
#include <fileio.h>
#include <libpad.h>
#include <loadfile.h>
#include <iopcontrol.h>
#include <iopheap.h>
#include <libmc.h>

#include "irx/cdvd_irx.h"
#include "irx/sjpcm_irx.h"

#include "cdvd_rpc.h"
#include "sjpcm.h"
#include "ps2.h"

#include "g2lib/g2.h"
#include "vmsave.h"


#ifdef main
#undef main
#endif

const char *gGameName;
const char *gGameNameShort;
const bool *gPlaybackStopped;
int mcard_present;
int progress = -1;

OSystem *OSystem_Ps2_create() {
	return OSystem_Ps2::create();
}

OSystem *OSystem_Ps2::create() {
	OSystem_Ps2 *syst = new OSystem_Ps2();
	return syst;
}

OSystem_Ps2::OSystem_Ps2()
  : screen(NULL), mouse(NULL), overlay(NULL), _ms_buf(NULL),
    _sound_proc(NULL), _timer_active(false), _current_shake_pos(0)
{
  char* ptr;

  msecs = 0;
  oldMsecs = 0;
  _active_buffer = 0;
  _avail_frames = 10;
  _playback_stopped = true;
  _devpoll = 0;
  gPlaybackStopped = &_playback_stopped;
  mouse_tx = NULL;
  backg_tx = NULL;
  screen = NULL;
  overlay = NULL;
  _ms_last_x = _ms_last_y = _ms_last_w = _ms_last_h = 0;
  _ps2_sound_buffer = 2;
  _top_screen = 24;

//  short temp_sound_buffer[2][RING_BUFFER_SAMPLES] __attribute__((aligned(64)));
  ptr = (char*) malloc(RING_BUFFER_SAMPLES * 2 + 64);
  temp_sound_buffer_ptr[0]  = (short*) ptr;
  temp_sound_buffer[0]  = (short*) (ptr + (64-((unsigned int)ptr % 64)));
  memset(temp_sound_buffer[0], 0, RING_BUFFER_SAMPLES * 2);


  ptr = (char*) malloc(RING_BUFFER_SAMPLES * 2 + 64);
  temp_sound_buffer_ptr[1]  = (short*) ptr;
  temp_sound_buffer[1]  = (short*) (ptr + (64-((unsigned int)ptr % 64)));
  memset(temp_sound_buffer[1], 0, RING_BUFFER_SAMPLES * 2);
}


/* CD Audio */
bool OSystem_Ps2::openCD(int drive)
{
  // Dummy.
  return true;
}

void OSystem_Ps2::playCD(int track, int num_loops, int start_frame, int duration)
{
}

void OSystem_Ps2::stopCD()
{
}

bool OSystem_Ps2::pollCD()
{
 return false;
}

void OSystem_Ps2::updateCD()
{
  // Dummy.  The CD drive takes care of itself.
}

void OSystem_Ps2::quit() {
  exit(0);
}

void OSystem_Ps2::setWindowCaption(const char *caption)
{
  gGameName = caption;
}

/* Mutex handling */
OSystem::MutexRef OSystem_Ps2::createMutex()
{
  return NULL;
}

void OSystem_Ps2::lockMutex(MutexRef mutex)
{
}
 
void OSystem_Ps2::unlockMutex(MutexRef mutex)
{
}

void OSystem_Ps2::deleteMutex(MutexRef mutex)
{
}


/* Features */
bool OSystem_Ps2::hasFeature(Feature f)
{

  switch(f) {
/*
  case kFeatureAspectRatioCorrection:
  case kFeatureVirtualKeyboard:
    return true;
*/
  case kFeatureLargeSoundBuffer:
	return true;
  default:
    return false;
  }

}

void OSystem_Ps2::setFeatureState(Feature f, bool enable)
{

  switch(f) {
/*
  case kFeatureAspectRatioCorrection:
    _aspect_stretch = enable;
    if(screen)
      setScaling();
    break;
  case kFeatureVirtualKeyboard:
    _softkbd_on = enable;
    break;
*/
  case kFeatureLargeSoundBuffer:
	if (enable) {
		_ps2_sound_buffer = 5;
	} else {
		_ps2_sound_buffer = 2;
	}
	break;
  default:
    break;
  }

}

bool OSystem_Ps2::getFeatureState(Feature f)
{

  switch(f) {
/*
  case kFeatureAspectRatioCorrection:
    return _aspect_stretch;
  case kFeatureVirtualKeyboard:
    return _softkbd_on;
*/
  default:
  case kFeatureLargeSoundBuffer:
	if (_ps2_sound_buffer > 2) {
		return true;
	} else {
		return false;
	}
	break;
  }
  return false;
}


int ps2_init_cdvd(const char * path1, const char * path2) {
    int ret;
    char moduleName[64];

    strcpy(moduleName, path1);
    strcat(moduleName, "CDVD.IRX");
    if (path1[0] == 'c' &&  path1[2] == 'r') strcat(moduleName, ";1");

    ret = SifLoadModule(moduleName, 0,NULL);
    if (ret < 0) {
	strcpy(moduleName, path2);
	strcat(moduleName, "CDVD.IRX");
        if (path2[0] == 'c' &&  path2[2] == 'r') strcat(moduleName, ";1");
	ret = SifLoadModule(moduleName, 0,NULL);
	if (ret < 0) {
		scr_printf("Load module CDVD.IRX failed. Load paths: %s, %s\n", path1, path2);
		printf("sifLoadModule %s failed: %d\n", moduleName, ret);
		return -1;
	}
    }
    ret = 0x01000000;
    while(ret--) asm("nop\nnop\nnop\nnop");

    CDVD_Init();
    return 1;
}

/* usb mass storage */

int check_usbd(const char * loadPath, const char * loadPathAlt) {
	int fd;
	char moduleName[64];
	int size;

	strcpy(moduleName, loadPath);
	strcat(moduleName, "USBD.IRX");

	fd = fioOpen(moduleName, O_RDONLY); 
	if (fd < 0) {
		strcpy(moduleName, loadPathAlt);
		strcat(moduleName, "USBD.IRX");
		fd = fioOpen(moduleName, O_RDONLY); 
	}


	if (fd < 0) {
		return 0;
	}
	
	size = fioLseek(fd, 0, SEEK_END);
	fioClose(fd);

	//Most of the newer usbd.irx'es are incomatible with ps2dev and cause
	//ps2 to freeze. Older versions (up to version 0.17.0) seems to be ok.
	if (size > 32760) {
		return -100;
	}
	return 1;
}

void dump_usbd_warning() {
	int i;

	scr_printf("\n \n \n");
	scr_printf("     !NOTE! \n\n\n");
	scr_printf("     USBD.IRX was found but unfortunately \n\n");
	scr_printf("     is not compatible. Please replace it \n\n");
	scr_printf("     by the older version (up to 0.17.0) or \n\n");
	scr_printf("     by the NPM-USBD.IRX \n\n");
	scr_printf("     USB functionality is now disabled. \n\n");

	for (i = 0; i <0xFFF * 20000; i++) {
	    asm("nop\nnop\nnop\nnop");
	}
}

int ps2_init_fs(const char * path1, const char * path2) {
    int ret;
    char moduleName[64];

    ret = check_usbd(path1, path2);    
    if (ret < 0) {
	return ret;
    }

    strcpy(moduleName, path1);
    strcat(moduleName, "USBD.IRX");
    if (path1[0] == 'c' &&  path1[2] == 'r') strcat(moduleName, ";1");
    ret = SifLoadModule(moduleName, 0,NULL);

    if (ret < 0) {
	strcpy(moduleName, path2);
	strcat(moduleName, "USBD.IRX");
        if (path2[0] == 'c' &&  path2[2] == 'r') strcat(moduleName, ";1");
	ret = SifLoadModule(moduleName, 0,NULL);

	if (ret < 0) {
		strcpy(moduleName, path1);
		strcat(moduleName, "NPM-USBD.IRX");
                if (path2[0] == 'c' &&  path2[2] == 'r') strcat(moduleName, ";1");
		ret = SifLoadModule(moduleName, 0,NULL);
		if (ret < 0) {
			strcpy(moduleName, path2);
			strcat(moduleName, "NPM-USBD.IRX");
                        if (path2[0] == 'c' &&  path2[2] == 'r') strcat(moduleName, ";1");
			ret = SifLoadModule(moduleName, 0,NULL);
		}
	}
    }
    if (ret < 0) {
	printf("sifLoadModule USBD.IRX failed: %d\n", ret);
	return -1;
    } 

    strcpy(moduleName, path1);
    strcat(moduleName, "USB_MASS.IRX");
    if (path1[0] == 'c' &&  path1[2] == 'r') strcat(moduleName, ";1");

    ret = SifLoadModule(moduleName, 0,NULL);
    if (ret < 0) {
	strcpy(moduleName, path2);
	strcat(moduleName, "USB_MASS.IRX");
        if (path2[0] == 'c' &&  path2[2] == 'r') strcat(moduleName, ";1");
	ret = SifLoadModule(moduleName, 0,NULL);
	if (ret < 0) {
		printf("sifLoadModule %s failed: %d\n", moduleName, ret);
	}
    }
    return ret;
}

void ps2_reset_iop() {

	SifIopReset("rom0:UDNL rom0:EELOADCNF",0);
	while (SifIopSync()) ;
	fioExit();
	SifExitIopHeap();
	SifLoadFileExit();
	SifExitRpc();
	SifExitCmd();
	EI;
	SifInitRpc(0);
	FlushCache(0);
	FlushCache(2);
}

void ps2_load_rom_modules() {
	int ret;

	ret = SifLoadModule("rom0:SIO2MAN", 0, NULL);
	if (ret < 0) {
		scr_printf("Failed to load module: SIO2MAN");
		while(1);
	}
	ret = SifLoadModule("rom0:PADMAN", 0, NULL);
	if (ret < 0) {
	        printf("sifLoadModule pad failed: %d\n", ret);
		while(1);
	}

	ret = SifLoadModule("rom0:MCMAN", 0, NULL);
	if (ret < 0) {
		scr_printf("Failed to load module: MCMAN");
		while(1);
	}

	ret = SifLoadModule("rom0:MCSERV", 0, NULL);
	if (ret < 0) {
		scr_printf("Failed to load module: MCSERV");
		while(1);
	}

	if(mcInit(MC_TYPE_MC) < 0) {
		printf("Failed to initialise memcard server!\n");
		while(1);
	}

	ret = SifLoadModule("rom0:LIBSD", 0,NULL);
	if (ret < 0) {
		scr_printf("sifLoadModule SD failed: %d\n", ret);
		while(1);
	}
}

void ps2_load_embeded_modules( int mode) {
	int ret;
	int iop_ret;

	ret = 0; 
	iop_ret = 0;
	scr_printf("loading embeded modules... \n");
	ret = SifExecModuleBuffer(cdvd_irx, CDVD_IRX_SIZE, 0, NULL, &iop_ret);
	if (ret < 0) {
		scr_printf("Exec module buffer: cdvd_irx failed: %d\n", ret);
		while(1);
	} else scr_printf("cdvd_irx ok! \n");
	//short delay 
	ret = 0x01000000;
	while(ret--) asm("nop\nnop\nnop\nnop");
	CDVD_Init();


	ret = 0; 
	iop_ret = 0;
	ret = SifExecModuleBuffer(sjpcm_irx, SJPCM_IRX_SIZE, 0, NULL, &iop_ret);
	if (ret < 0) {
		scr_printf("Exec module buffer: sjpcm_irx failed: %d\n", ret);
		while(1);
	} else scr_printf("sjpcm_irx ok! \n");
	//short delay 
	ret = 0x01000000;
	while(ret--) asm("nop\nnop\nnop\nnop");

	 //non-sync mode
	SjPCM_Init(0, mode);
	SjPCM_Setvol(0x3fff);
}


int ps2_check_modules_available(const char * path1, const char * path2) {
    int ret;
    char moduleName[64];

    strcpy(moduleName, path1);
    strcat(moduleName, "CDVD.IRX");
    if (path1[0] == 'c' && path1[2] == 'r') { //cdrom
        strcat(moduleName, ";1");
    }

    ret = fioOpen(moduleName, O_RDONLY);

    if (ret < 0) {
	strcpy(moduleName, path2);
	strcat(moduleName, "CDVD.IRX");
        if (path2[0] == 'c' && path2[2] == 'r') { //cdrom
            strcat(moduleName, ";1");
        }
	ret = fioOpen(moduleName, O_RDONLY);
    }

    if (ret >0) {
        fioClose(ret);
    }
    return ret;
}



#define PS2_PATH1 "cdrom0:\\"
#define PS2_PATH2 "host0:"
#define PS2_PATH3 "cdfs:\\"

void ps2_init_hardware(void)
{
	int ret;
	int modules_available;
	int pal_mode;
	pal_mode = 0;

	SifInitRpc(0); 
	init_scr();
	pal_mode = g2_is_pal();

	modules_available = ps2_check_modules_available(PS2_PATH1, PS2_PATH2);

	//machine has NO access to modules - reset iop and load embeded modules
	if (modules_available < 0 ) {
		printf("cdvd.irx not found - forcing load of embeded modules... \n");
		ps2_reset_iop();
		ps2_load_rom_modules();
		ps2_init_pad();  
		ps2_load_embeded_modules(pal_mode);
		ps2_init_sound(NULL, NULL, pal_mode);
		ret = ps2_init_fs(PS2_PATH1, PS2_PATH3); //other filesystems
	} else {
	//modules are avaialble
		ps2_load_rom_modules();
		ps2_init_pad();  
   		ps2_init_cdvd(PS2_PATH1, PS2_PATH2);
		ps2_init_sound(PS2_PATH1, PS2_PATH2, pal_mode);
		ret = ps2_init_fs(PS2_PATH2, PS2_PATH3); //other filesystems
	}
	if (ret == -100) {
		dump_usbd_warning();
	}
#ifndef PCSX2
	mcard_present = ps2_init_mcard(); 
#else
	mcard_present = 1;
#endif
   
	//set videomode   
	if (pal_mode) {
		g2_init(PAL_320_256_32);
	} else {
		g2_init(NTSC_320_224_32);
	}
	// clear the screen
	g2_set_fill_color(0, 0, 0);
	g2_fill_rect(0, 0, g2_get_max_x(), g2_get_max_y());

}



int ps2_check_local_dir(const char* gameName ) {

   if (mcard_present < 0) {
	printf ("*  No memory card found in slot 0 \n");
	return -1;
   }
   return ps2_check_mcard_root_dir(gameName) ;
}

extern "C" int scummvm_main(int argc, char *argv[]);
int main()
{
  
  char gname[32];
  //char inipath[128];
  char* inipath;
  char* gamedir;
  char gameDirParam[256];   
  static char *argv[] = { "scummvm", "-n", gamedir, gname};
  static int argc = 4;

  printf("SCUMMVM start.\n");
  ps2_init_hardware();

  if(!selectGame(argv[3], inipath, argv[2]))
    exit(0);

  gGameNameShort = argv[3];
#ifndef PCSX2
  ps2_check_local_dir(argv[3]); //check local dir on memory card
#endif

  gameDirParam[0] = 0;
  strcat(gameDirParam, "--path=");
  strcat(gameDirParam, argv[2]);
  argv[2] =(char*) &gameDirParam;
  scummvm_main(argc, argv);

  while(1);
}

