/* ScummVM - Scumm Interpreter
 * PS2 port
 * Copyright (C) 2002  Marcus Comstedt
 * For PS2 altered by Ole
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header: /cvsroot/scummvm/scummvm/backends/dc/selector.cpp,v 1.11.2.1 2003/08/05 20:01:04 marcus_c Exp $
 *
 */

#include <common/stdafx.h>
#include <common/scummsys.h>
#include <base/engine.h>
#include <base/gameDetector.h>
#include <base/plugins.h>
#include <backends/fs/fs.h>
#include <fileio.h>
#include <tamtypes.h>

#include "g2lib/g2.h"
#include "ps2.h"
#include "controls.h"
#include "cdvd_rpc.h"

#define MAX_GAMES 100


struct Game
{
  char dir[256];            //??? path to game ???
  char filename_base[256];  //??? short name   ???
  char text[256];           //??? description  ???
  //Icon icon;
  //Label label;
};


struct Dir
{
  char name[256];
  char deficon[256];
};


static const char * iniFiles[] = {
			     "mass:/scummvm.ini",
			     "host0:scummvm.ini",
			     "cdfs:/scummvm.ini",
                             "cdrom0:\\SCUMMVM.INI;1",
                             "cdrom0:\\DATA\\SCUMMVM.INI;1",
                             "mc0:\\SCUMMVM\\SCUMMVM.INI",
			     NULL
                           };
static int iniSelection;


static Game the_game;
//buffer for game labels (bitmaps: each label has 256x10 pixels = 2560 bytes)
unsigned char * labels;
//buffer for scummvm logo texture (112*16)
unsigned char * logo;


static bool checkName(const char *base, char *text = 0)
{
  GameDetector g;
  GameSettings gs = g.findGame(base);

  if (gs.name) {
    if(text != NULL)
      strcpy(text, gs.description);
    return true;
  }
  return false;
}

static const char *checkDetect(const FilesystemNode *entry, bool unique)
{
  FSList files;
  files.push_back(*entry);
  DetectedGameList candidates;

  const PluginList &plugins = PluginManager::instance().getPlugins();
  PluginList::const_iterator iter = plugins.begin();
  for (iter = plugins.begin(); iter != plugins.end(); ++iter) {
    candidates.push_back((*iter)->detectGames(files));
  }
  if (candidates.isEmpty())
    return NULL;
  if (unique && candidates.size() > 1)
    return NULL;
  return candidates[0].name;
}

static bool isGame(const FilesystemNode *entry, char *base)
{
  const char *fn = entry->displayName().c_str();
  if(const char *dtct = checkDetect(entry, true)) {
    strcpy(base, dtct);
    return true;
  }
  if(!strcasecmp(fn, "00.LFL") ||
     !strcasecmp(fn, "000.LFL")) {
    *base = '\0';
    return true;
  }
  if(const char *dtct = checkDetect(entry, false)) {
    strcpy(base, dtct);
    return true;
  }
#if 0
  int l = strlen(fn);
  if(l>4 && (!strcasecmp(fn+l-4, ".000") ||
	     !strcasecmp(fn+l-4, ".SM0") ||
	     !strcasecmp(fn+l-4, ".HE0") ||
	     !strcasecmp(fn+l-4, ".LA0"))) {
    strcpy(base, fn);
    base[l-4]='\0';
    return true;
  }
#elsif 0
  char *dot;
  if(!stricmp(fn, "LOOM.EXE"))
    return false;
  if((dot = strchr(fn, '.'))!=NULL) {
    if(!strcasecmp(dot, ".SAN"))
      return false;
    strcpy(base, fn);
    base[dot-fn]='\0';
    if(checkName(base))
      return true;
  }
#endif  
  return false;
}





static int parseConfig(char* buf, int bufSize,  Game *games, int max)
{
  char line[512]; //one line of the text file - limited to 512 characters
  int lineIndex;
  int i;
  int length = 0;
  int curr_game;

  i = 0;
  lineIndex = 0;
  curr_game = -1;

  while (i < bufSize) { //don't go beyond buffer boudaries

    if (buf[i] == 13) { //Windows Carriage return - ususaly at the end of line - we don't need it 
         buf[i]=0; 
    } else
    if (buf[i] == 10) { //Line feed - end of line - analyse line
        line[lineIndex] = 0; //terminate the line

        //remove trailing space and tabulators
        lineIndex--;
        while(line[lineIndex] == 32 || line[lineIndex] == 9) {
            line[lineIndex] = 0;
            lineIndex--;
        }
           
        length = strlen(line);
     
        //todo - we could rotate line left (remove leading spaces)

	//now analyse line
        if (line[0] == '[')  { //found a game headline
		curr_game++;
		strncpy(games[curr_game].filename_base, line+1, length-2); //skip first and last char
		games[curr_game].filename_base[length-2] = 0;
                games[curr_game].text[0] = 0;
        } else 
        if (strncmp(line,"path", 4) == 0){ //found a path
           lineIndex = 4;
           //skip unwanted characters
           while (line[lineIndex]==32 ||  		//space
                  line[lineIndex]== 9 ||  		//tabulator
                  line[lineIndex]=='=')   lineIndex++;
           //copy path
           strncpy(games[curr_game].dir, line+lineIndex, length-lineIndex); //skip first and last char
	   games[curr_game].dir[length-lineIndex] = 0;
          
        } else
        if (strncmp(line,"description", 11) == 0){ //found a description
           lineIndex = 11;
           //skip unwanted characters
           while (line[lineIndex]==32 ||  		//space
                  line[lineIndex]== 9 ||  		//tabulator
                  line[lineIndex]=='=')   lineIndex++;
           //copy description
           strncpy(games[curr_game].text, line+lineIndex, length-lineIndex); //skip first and last char
	   games[curr_game].text[length-lineIndex] = 0;

        } 
          
        lineIndex = 0; //prepare for next line
    } else { //add character to line
        line[lineIndex] = buf[i];
        lineIndex++;
    }

    i++;

  }
  curr_game++;

  for (i = 0;  i < curr_game; i++) {
    if (games[i].text[0] == 0) 
      strcpy(games[i].text, games[i].filename_base);
  }

  return curr_game;

}
static int findGames(Game *games, int max)
{

    int fd;
    int size;
    int i;
    char* buffer;



    //try to find scummvm.ini file
    iniSelection = 0;
    fd = -1;
    while (fd < 0 && iniFiles[iniSelection]!= NULL) {
	fd = fioOpen(iniFiles[iniSelection], O_RDONLY); 
	iniSelection++;
    }

    //config file was not found
    if (fd < 0) {
	return -1;
    }
    
    size = fioLseek(fd, 0, SEEK_END);
    fioLseek(fd,0,SEEK_SET);
    buffer = (char*)malloc(size+1);

    if (buffer == NULL) {
      printf ("Malloc failed \n");
      fioClose(fd);
      return -1;
    }                    
    buffer[size] = 0;


    i = fioRead(fd, buffer, size);
    if (i != size) {
       printf ("Error reading ini file! \n");
       size = i;
    }

   fioClose(fd);
 
   //parse config file
   i = parseConfig(buffer, size, games, max);
   free(buffer);
   CDVD_Stop();

   /**  debug */
/*
   for (fd = 0; fd < i ; fd ++) {
      printf ("[%s] %s path=%s \n", games[fd].filename_base, games[fd].text, games[fd].dir);
   }
*/
   return i;

}


static void createLogo() {
  int i, j;
  int idx;
  unsigned char* dst = logo;

  //no memory allocated?
  if (logo == NULL) {
    return;
  }

  for (j = 46; j < 62; j++) {     // 16 pixels high
    idx = j*320;
    for (i = 208; i < 320; i++) { //112 pixels width
       *dst = controls_img[idx+i-1];
       dst++;
    }
  }
}

static void createLabelTexture(int labelIndex, unsigned char* name) {

  int shift,i,j,k;
  unsigned char * dst; 
  unsigned char * src;
  int fntIdx = 0;
  int x,y;
  int length = strlen((const char*)name);
  
  if (length > 32) {
    length = 32;
  }

  // center names
  shift = (32-length) / 2;



  //clear label background;
  dst = labels + (2560*labelIndex);
  for (j = 0; j < 2560; j++) {
        *dst = 51;          
        dst++;
  }


  // copy font pixels to destination label texture
  for (i=0; i < length; i++) {  //go through each character
     fntIdx = font_idx[name[i]];
     y = fntIdx / 26;
     x = fntIdx % 26;
     src =  controls_img + 320*45 + y*320*9 + x*8;
     src--;     
     dst =  labels + (2560*labelIndex)+ 8*(shift+i);
     for (j = 0; j < 10; j++) { // character is 10 lines high

       for (k = 0; k < 8; k++) {// character is 8 pixels width 
          *dst = *src;
          dst++;
          src++;
       }
       src+=320-8;  //width of the font image - 8 pixels already drawn
       dst+=256-8;  //width of the label image - 8 pixels already drawn
     }
  }
}

void showMessage(char* name) {
  unsigned char pixmap[2560];
  int shift,i,j,k;
  unsigned char * dst; 
  unsigned char * src;
  int fntIdx = 0;
  int x,y;
  int length = strlen((const char*)name);
  
  if (length > 32) {
    length = 32;
  }

  // left offset
  shift = 0;

  //clear label background;
  dst = pixmap;
  for (j = 0; j < 2560; j++) {
        *dst = 51;          
        dst++;
  }


  // copy font pixels to destination label texture
  for (i=0; i < length; i++) {  //go through each character
     fntIdx = font_idx[name[i]];
     y = fntIdx / 26;
     x = fntIdx % 26;
     src =  controls_img + 320*45 + y*320*9 + x*8;
     src--;     
     dst =  pixmap + 8*(shift+i);
     for (j = 0; j < 10; j++) { // character is 10 lines high

       for (k = 0; k < 8; k++) {// character is 8 pixels width 
          *dst = *src;
          dst++;
          src++;
       }
       src+=320-8;  //width of the font image - 8 pixels already drawn
       dst+=256-8;  //width of the label image - 8 pixels already drawn
     }
  }
  g2_put_image8(32, 200, 256,10, pixmap, (unsigned int*)  controls_pal,1);
}

static void drawBackground()
{

 int offs = 51 * 4;

 g2_set_fill_color((unsigned int)controls_pal[offs], (unsigned int) controls_pal[offs+1], (unsigned int)controls_pal[offs+2]);
 g2_fill_rect(0,45+24,319,200+24);
 g2_put_image8(0,0+24,320, 45, controls_img, (unsigned int*)controls_pal, 1);

 if (logo != NULL) {
   g2_put_image8(104,50+24,112, 16, logo, (unsigned int*)controls_pal, 1);
 }
   
 g2_set_fill_color(0,0,0);
}

static void clearScreen() 
{
 g2_set_fill_color(0, 0 ,0);
 g2_fill_rect(0,24,319,200+24);
}

static void drawGameLabel(int idx, int x, int y, int lit)
{
    if (lit) 
     g2_put_image8(x, y+24, 256,10, labels+(2560*idx), ((unsigned int*) controls_pal)+3, 1);//+2
    else 
     g2_put_image8(x, y+24, 256,10, labels+(2560*idx), (unsigned int*)controls_pal,1);
}

int gameMenu(Game *games, int num_games)
{
  int top_game = 0, selector_pos = 0;
  int mousex = 0, mousey = 0;
  byte shiftFlags;
  int y;
  int i, cnt;
  int event = 1;


  //no games detected 
  if(!num_games)
    return -1;

  //draw background 
  drawBackground();


  for(;;) {

    //Redraw labels if key pressed
    if (event) {
	    y = 80;
	    g2_wait_vsync();
	    for(i=top_game, cnt=0; cnt<7 && i<num_games; i++, cnt++) {
	      drawGameLabel(i, 32, y, (cnt == selector_pos? 1 : 0));
	      y += 14;
	    }
	    //key repeat delay
	    for (i = 0; i < 10; i++) g2_wait_vsync();
    }


    event = handleInput(&mousex, &mousey, shiftFlags,1,24);

    //Enter pressed or LMB or START    
    if(event==-OSystem::EVENT_LBUTTONDOWN ||  event==319) {
      int selected_game = top_game + selector_pos;
      clearScreen();
      return selected_game;
    }else
    //user wants to refresh 
    if (event == 13) {
	return -1;
    }
	

    //Key Down pressed (for timing reason the treshold is 16)
    if(mousey>16) {
      if(selector_pos + top_game + 1 < num_games)
	if(++selector_pos >= 7) {
	  --selector_pos;
	  ++top_game;
          
	}
      mousey = 0;
      event = 1;

    } else
    //Key up pressed
    if(mousey<=-16) {
      if(selector_pos + top_game > 0)
	if(--selector_pos < 0) {
	  ++selector_pos;
	  --top_game;
	}
      mousey = 0;
      event = 1;
      
    }

  }
}

void waitForDisk(Game* games, int &num_games)
{
	if (num_games > 0) {
	   free(labels);
	}
	CDVD_FlushCache();
	//Get available games info from current cd-rom
	num_games = findGames(games, MAX_GAMES);
	if (num_games > 0) {
		labels = (unsigned char*) malloc(2560 * num_games);
		//Create labels of available games
		for(int i=0; i<num_games; i++) {
			createLabelTexture(i, (unsigned char*) games[i].text);
		}

	}
}


bool selectGame(char *&ret, char*& ini, char*& pth)
{
  Game *games = new Game[MAX_GAMES];
  int selected, num_games;

  num_games = 0;
  //Create buffer for logo
  logo   = (unsigned char*) malloc(1792);

  //Create logo texture
  createLogo();

  
  for(;;) {
    waitForDisk(games, num_games);

    //Wait until user pickups the item
    selected = gameMenu(games, num_games);

    //Selection made 
    if(selected >= 0)
      break;

  }
/*
  //Error checking
  if(selected >= num_games)
    selected = -1;
*/
  //store selected game info
  if(selected >= 0) {
	   strcpy(the_game.filename_base, games[selected].filename_base);
	   strcpy(the_game.dir, games[selected].dir);
	   
  }

  printf("Selected=%i game=%s  dir=%s \n", selected, the_game.filename_base, the_game.dir);
  //free memory
  free (logo);
  if (num_games > 0)  {
	  free (labels);
  }
  delete games;

  //set arguments  and exit
  if(selected>=0) {


    ini = (char*)iniFiles[iniSelection];
    ret = the_game.filename_base;
    pth = the_game.dir;

    return true;
  } else
    return false;
}
