/* ScummVM - Scumm Interpreter
 * PS2 port
 * Copyright (C) 2002  Marcus Comstedt
 * For PS2 altered by Ole
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header: /cvsroot/scummvm/scummvm/backends/dc/time.cpp,v 1.3 2003/06/18 21:15:08 marcus_c Exp $
 *
 */

#include <common/stdafx.h>
#include <common/scummsys.h>
#include <kernel.h>
#include <sifrpc.h>
#include <loadfile.h>
#include "ps2.h"
#include "g2lib/g2.h"
#include <libpad.h>

#define PAL_FRAMES 50
#define	TIMER_IRX	0x8000012A

static SifRpcClientData_t client __attribute__((aligned(64)));
unsigned char rpcBuffer[128] __attribute__((aligned(64)));

#ifndef PCSX2

uint32 OSystem_Ps2::getMillis()
{
  uint32 x;
  asm("mfc0\t%0, $9\n" : "=r"(x));   //get_ticks
  x /= 290000; //333Mhz machine
  if (x < oldMsecs) {
	msecs += (14810 - oldMsecs);  //14316-300  13766-312  14810-290
        msecs += x;
  } else {
        msecs +=  (x-oldMsecs);
  }
  oldMsecs = x;
  return msecs;
}

void OSystem_Ps2::delayMillis(uint msecs_time)
{
 uint32 nowTime = getMillis();
 uint32 finalTime = nowTime + msecs_time;

 while (nowTime <= finalTime) {
   checkSound();
   nowTime= getMillis();    
 }
}

/*
uint32 OSystem_Ps2::getMillis()
{
  msecs+= 5;
  return msecs;
}


void OSystem_Ps2::delayMillis(uint msecs_time)
{
 int i;

  for (i = 0; i <0xFFF * msecs_time; i++) {
    if (i % 512 == 0) checkSound();
    asm("nop\nnop\nnop\nnop");
//      checkSound();
  }

}
*/
#else /* PCSX2*/



uint32 OSystem_Ps2::getMillis()
{
   int* sec;
   int* usec;

   SifCallRpc(&client,0,0,(void*)(&rpcBuffer[0]),0,(void*)(&rpcBuffer[0]),128,0,0);
   sec = (int*) (&rpcBuffer[0]);
   usec = (int*) (&rpcBuffer[4]);
   msecs = (*sec)*1000+ (*usec)/1000;
   return msecs;
}



void OSystem_Ps2::delayMillis(uint msecs_time)
{
 uint32 nowTime = getMillis();
 uint32 finalTime = nowTime + msecs_time;

 while (nowTime <= finalTime) {
   checkSound();
   nowTime= getMillis();    
 }
 
}

#endif /* PCSX2*/

void OSystem_Ps2::setTimerCallback(TimerProc callback, int timer)
{
    if (callback != NULL) {

      _timer_duration = timer;
      _timer_next_expiry = getMillis() + timer;
      _timer_callback = callback;
      _timer_active = true;

       //printf("Setting timer to :%i \n", timer);
    } else {
      _timer_active = false;
    }



}

void ps2_init_timer(const char * loadPath, const char * loadPathAlt)
{
    int ret;
    char moduleName[64];
    strcpy(moduleName, loadPath);
    strcat(moduleName, "TIMER.IRX");

    ret = SifLoadModule(moduleName,0, NULL);
    if (ret < 0) {
	strcpy(moduleName, loadPathAlt);
	strcat(moduleName, "TIMER.IRX");
	ret = SifLoadModule(moduleName,0, NULL);
	if (ret < 0) {
		scr_printf("Load module TIMER.IRX failed. Load paths: %s, %s\n", loadPath, loadPathAlt);
        	printf("sifLoadModule %s failed: %d\n", moduleName, ret);
        	while(1);
	}
    }


    //short delay - give the module enough time to register itself
    ret = 0x01000000;
    while(ret--) asm("nop\nnop\nnop\nnop");


    //printf ("Trying to bind Timer...\n");
    while (1) {
        ret = SifBindRpc( &client, TIMER_IRX, 0);
        if ( ret  < 0)  {
           break;
        }
        if (client.server != 0) break;

        // short delay ???
      	ret = 0x10000;
    	while(ret--);

    }
    if (ret < 0 ) {
        printf("SifBindRpc Timer failed: %d !!!!\n", ret);
    }

}
