/* ScummVM - Scumm Interpreter
 * PS2 port
 * Copyright (C) 2002  Marcus Comstedt
 * For PS2 altered by Ole
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header: /cvsroot/scummvm/scummvm/backends/dc/vmsave.cpp,v 1.8 2003/06/18 21:15:08 marcus_c Exp $
 *
 */

#include <common/stdafx.h>
#include <common/scummsys.h>
#include <base/engine.h>
#include <scumm/saveload.h>

#include <gui/newgui.h>
#include <gui/message.h>

#include <sifrpc.h>
#include <loadfile.h>
#include <fileio.h>
#include <libmc.h>

#include "ps2.h"
#include "audio.h"
#include "icon.h"
#include "vmsave.h"
#include "audio.h"

#define TYPE_MC
//#define TYPE_XMC

// Savegame can not be bigger than this, even before compression
#define MAX_SAVE_SIZE (128*1024)

#define ARRAY_ENTRIES	64

enum vmsaveResult {
  VMSAVE_OK,
  VMSAVE_NOVM,
  VMSAVE_NOSPACE,
  VMSAVE_WRITEERROR,  
};


static mcTable mcDir[ARRAY_ENTRIES] __attribute__((aligned(64)));
static int mc_Type, mc_Free, mc_Format;
static int _entry_count;



static void displaySaveResult(vmsaveResult res)
{
  char buf[1024];

  switch(res) {
  case VMSAVE_OK:
//    sprintf(buf, "Game saved !");
    return;
  case VMSAVE_NOVM:
    strcpy(buf, "No memory card present!");
    break;
  case VMSAVE_NOSPACE:
    strcpy(buf, "Not enough space available!");
    break;
  case VMSAVE_WRITEERROR:
    strcpy(buf, "Write error!!!");
    break;
  default:
    strcpy(buf, "Unknown error!!!");
    break;
  }

#ifndef PCSX2
  GUI::MessageDialog dialog(buf);
  dialog.runModal();
#endif
}

/*
static void vmdelay(int count) {
  count*= 0xFFFF;
  while (count> 0)  {
     asm("\nnop\nnop\nnop\nnop");
     count--;
  }
}
*/

static vmsaveResult trySave(const char *filename, const char *data, int size,
			    int port)
{
  int fd;
  int ret;

  printf ("**trySave start filename=%s size=%i \n", filename, size);

  pause_sound(); //sound starts to play automatically

/*
  if(!ps2_check_mcard_unit(port, 0))
    return VMSAVE_NOVM;

  if(ps2_mcard_free_space(port,0) < size) 
	return VMSAVE_NOSPACE;

*/
  fd = fioOpen(filename, O_WRONLY | O_CREAT);
  if (fd <= 0)  {
        printf ("**trySave failed during fioOpen filename=%s size=%i \n", filename, size);

	return VMSAVE_WRITEERROR;
  }

  ret = fioWrite(fd, (char*) data, size);

  fioClose(fd);
  if (ret!= size) {
    printf ("**trySave failed during fioWrite filename=%s size=%i \n", filename, size);
    return VMSAVE_WRITEERROR;
  }
  return VMSAVE_OK;
}

static bool tryLoad(char *&buffer, int &size, const char *filename, int port)
{
  int fd;
  int fileSize;
  int ret;

  pause_sound(); //sound starts to play automatically

  printf ("**tryLoad start filename=%s size=%i buffer=%i \n", filename, size, (unsigned int) buffer);

/*
  if(!ps2_check_mcard_unit(port, 0))
    return false;
*/

  fd = fioOpen(filename, O_RDONLY);
  if(fd< 0) {
    printf ("**tryLoad mcOpen failed filename=%s size=%i \n", filename, size);
    return false;
  }
  fileSize = fioLseek(fd, 0, SEEK_END);
  fioLseek(fd, 0, SEEK_SET);

  size=fileSize;
  printf("** tryLoad real size=%d \n",size);

  buffer = NULL;
  buffer = new char[fileSize];


  ret = fioRead(fd, buffer, size);
  fioClose(fd);

  if (ret <=  0){
     printf ("**tryLoad failed filename=%s size=%i \n", filename, size);
     if (buffer != NULL) 
	     delete buffer;
     return false;
   }  else {
  //read succesfull
     return true;
  }

}


static void clearEntries() {
 int i;
 for (i = 0; i < ARRAY_ENTRIES; i++) {
    mcDir[i].fileSizeByte = 0;
    mcDir[i].attrFile = 0;
    mcDir[i].unknown3 = 0;
    mcDir[i].name[0] = 0;
    mcDir[i].unknown4[0] = 0;
    mcDir[i].unknown4[1] = 0;
    mcDir[i]._create.sec = 0;
    mcDir[i]._create.min = 0;
    mcDir[i]._create.hour = 0;
    mcDir[i]._create.day = 0;
    mcDir[i]._create.month = 0;
    mcDir[i]._create.year = 0;
    mcDir[i]._create.unknown1 = 0;

    mcDir[i]._modify.unknown2 = 0;
    mcDir[i]._modify.sec = 0;
    mcDir[i]._modify.min = 0;
    mcDir[i]._modify.hour = 0;
    mcDir[i]._modify.day = 0;
    mcDir[i]._modify.month = 0;
    mcDir[i]._modify.year = 0;
 }
}

static void tryList(const char *prefix, bool *marks, int num, int port)
{
    extern const char *gGameNameShort;

    char ps2_path[128];
    strcpy(ps2_path, "/SCUMMVM-");
    strcat(ps2_path, gGameNameShort);
    strcat(ps2_path, "/");
    strcat(ps2_path, prefix);
    strcat(ps2_path, "*");

    int j, l, ret;
    long i = 42;
    int pl = strlen(prefix);
//
    clearEntries();

    mcGetDir(0, 0, ps2_path, 0, ARRAY_ENTRIES - 10, mcDir);
    mcSync(0, NULL, &ret);
 //
    printf("** Try List prefix=%s path=%s count=%i\n", prefix, ps2_path, _entry_count);
    for (j = 0; j < _entry_count; j++) {
	if(!(mcDir[j].attrFile & MC_ATTR_SUBDIR)) { //entry is not  directory
		printf(" *   %s\n", mcDir[j].name);
		char* buf =  (char*)mcDir[j].name;
		char* endp = NULL;
		l= strlen(buf);
		i = 42;
		if(l > pl && !strncmp(buf, prefix, pl) &&
 			(i = strtol(buf+pl, &endp, 10))>=0 && i<num &&
			(endp - buf) == l)
		  marks[i] = true;
	}	
    }



}

vmsaveResult writeSaveGame(const char *gamename, const char *data, int size,
			   const char *filename )
{
    char ps2_name[128];
    strcpy(ps2_name, "mc0:SCUMMVM-");
    strcat(ps2_name, gamename);
    strcat(ps2_name, "/");
    strcat(ps2_name, filename);

    //printf("** Try Save gamename=%s filename=%s \n", gamename, ps2_name);
    
    return trySave(ps2_name, data, size, 0); // port 0 only
}

int getSaveGameSize(const char* filename) {
  int size = -1;
  int i;
  for (i = 0; i < _entry_count; i++) {
    if (strcmp((char*)mcDir[i].name, filename) == 0)
 	 return mcDir[i].fileSizeByte;
  }
  
  return size;

}

bool readSaveGame(char *&buffer, int &size, const char *filename)
{

    extern const char *gGameNameShort;

    char ps2_name[128];
    strcpy(ps2_name, "mc0:SCUMMVM-");
//    strcpy(ps2_name, "/SCUMMVM-");
    strcat(ps2_name, gGameNameShort);
    strcat(ps2_name, "/");
    strcat(ps2_name, filename);

/*
    size = getSaveGameSize(filename);
    if (size < 0) { 
        printf("** readSaveGame error path=%s filename=%s entryCount=%i \n", ps2_name, filename, _entry_count);

	return false;
    }
*/

    return tryLoad(buffer, size, ps2_name, 0); // port 0 only
}



//Memory Card Save

class VMSave : public SaveFile {
private:

  bool issave;
  char *buffer;
  int pos, size;
  char filename[16];

  uint32 read(void *buf, uint32 cnt);
  uint32 write(const void *buf, uint32 cnt);

public:
  VMSave(const char *_filename, bool _saveOrLoad) 
    : issave(_saveOrLoad), pos(0), buffer(NULL)
  {
    strncpy(filename, _filename, 16);
    if(issave)
      buffer = new char[size = MAX_SAVE_SIZE];
  }


  ~VMSave();

  bool isOpen() const { return true; }
  bool readSaveGame()
  { return ::readSaveGame(buffer, size, filename); }

};

//Save Manager

class VMSaveManager : public SaveFileManager {
public:
  VMSaveManager(void);
  virtual SaveFile *openSavefile(const char *filename, bool saveOrLoad);
  virtual void listSavefiles(const char *prefix, bool *marks, int num);
};


//**********************************************************
//Implementation


//***[VMSave]*********************
VMSave::~VMSave()
{

  extern const char *gGameNameShort;

  if(issave) {
    displaySaveResult(writeSaveGame(gGameNameShort, buffer,
				    pos, filename));
  }
  if (buffer != NULL)
    delete buffer;

};

uint32 VMSave::write(const void *buf, uint32 cnt) {
  if (!issave)
    return 0;

  int nbyt = cnt;
  if (pos + nbyt > size) {
    cnt = (size - pos);
    nbyt = cnt;
  }
  if (nbyt)
    memcpy(buffer + pos, buf, nbyt);
  pos += nbyt;
  return cnt;

}
uint32 VMSave::read(void *buf, uint32 cnt)
{
  if (issave)
    return 0; 

  int nbyt = cnt;
  if (pos + nbyt > size) {
    cnt = (size - pos);
    nbyt = cnt;
  }
  if (nbyt)
    memcpy(buf, buffer + pos, nbyt);
  pos += nbyt;
  return cnt;
}






//***[Save Manager]*********************
VMSaveManager::VMSaveManager(void) {
	ps2_get_mcard_dir(0);  //port 0 only

	printf ("** VMSaveManager:get dir _entry_count=%i \n", _entry_count);

	if (_entry_count < 0) {
		while(1);
	}
}


SaveFile *VMSaveManager::openSavefile(const char *filename, bool saveOrLoad)
{

  VMSave *s = new VMSave(filename, saveOrLoad);
  if(saveOrLoad) //try save
    return s;
  else if(s->readSaveGame()) { //tryLoad
    return s;
  } else {
    delete s;
    return NULL;
  }

}

void VMSaveManager::listSavefiles(const char *prefix, bool *marks, int num)
{
  //clear info about available saveGameFiles
  memset(marks, false, num*sizeof(bool));
  tryList(prefix, marks, num, 0);

};

SaveFileManager *OSystem_Ps2::getSavefileManager()
{
  return new VMSaveManager();
}






// PS2 "low" level stuff

int  ps2_init_mcard(void ) {
	int ret;
	ret = 1;

	//Try detect inserted card - in port 0

	// Since this is the first call, -1 should be returned.
	mcGetInfo(0, 0, &mc_Type, &mc_Free, &mc_Format); 
	mcSync(0, NULL, &ret);
	printf("mcGetInfo returned %d\n",ret);
	printf("Type: %d Free: %d Format: %d\n\n", mc_Type, mc_Free, mc_Format);

	if (ret == -1) {  //switched to a formated card - just assure this
  		// Assuming that the same memory card is connected, this should return 0
		mcGetInfo(0,0,&mc_Type,&mc_Free,&mc_Format);
		mcSync(0, NULL, &ret);
		printf("mcGetInfo returned %d\n",ret);
		printf("Type: %d Free: %d Format: %d\n\n", mc_Type, mc_Free, mc_Format);
	}

	return ret;
}

int ps2_check_mcard_root_dir(const char* gameName) {
	int ret = -1;
	int fd;

        char pathName[128];
        char fileName[128];

        strcpy(pathName, "mc0:SCUMMVM-");
	strcat(pathName, gameName);

	strcpy(fileName, pathName);
	strcat(fileName, "/icon.sys"); //mandatory file for each root directory


	// Check if existing save is present
	fd = fioOpen(fileName, O_RDONLY);
	if(fd < 0) {

		printf("\nNo previous save exists, creating...\n");

		if((ret = ps2_create_mcard_root_dir(pathName, fileName, gameName)) < 0) {
			printf("Failed to create root dir %s ! Errorno: %d\n", fileName, ret);
			while(1);
		}

	} else {
		fioClose(fd);
	}
	return ret;
}

int ps2_create_mcard_root_dir(char* pathName, char * fileName, const char* gameName) {

	int mc_fd;
	int icon_fd;
	mcIcon icon_sys;
	char name[128];

	static iconIVECTOR bgcolor[4] = {
		{  68,  23, 116,  0 }, // top left
		{ 255, 255, 255,  0 }, // top right
		{ 255, 255, 255,  0 }, // bottom left
		{  68,  23, 116,  0 }, // bottom right
	};

	static iconFVECTOR lightdir[3] = {
		{ 0.5, 0.5, 0.5, 0.0 },
		{ 0.0,-0.4,-0.1, 0.0 },
		{-0.5,-0.5, 0.5, 0.0 },
	};

	static iconFVECTOR lightcol[3] = {
		{ 0.3, 0.3, 0.3, 0.00 },
		{ 0.4, 0.4, 0.4, 0.00 },
		{ 0.5, 0.5, 0.5, 0.00 },
	};

	static iconFVECTOR ambient = { 0.50, 0.50, 0.50, 0.00 };

	//Create directory
	if(fioMkdir(pathName) < 0) return -1;

	// Set up icon.sys. This is the file which controls how our memory card save looks
	// in the PS2 browser screen. It contains info on the bg colour, lighting, save name
	// and icon filenames. Please note that the save name is sjis encoded.
	strcpy(name, "SCUMMVM\n");
	strcat(name, gameName);

	memset(&icon_sys, 0, sizeof(mcIcon));
	strcpy((char*)icon_sys.head, "PS2D");
	strcpy_sjis((short *)&icon_sys.title, name);
	icon_sys.nlOffset = 16;
	icon_sys.trans = 0x60;
	memcpy(icon_sys.bgCol, bgcolor, sizeof(bgcolor));
	memcpy(icon_sys.lightDir, lightdir, sizeof(lightdir));
	memcpy(icon_sys.lightCol, lightcol, sizeof(lightcol));
	memcpy(icon_sys.lightAmbient, ambient, sizeof(ambient));
	strcpy((char*)icon_sys.view, "scummvm.icn"); // these filenames are relative to the directory
	strcpy((char*)icon_sys.copy, "scummvm.icn"); // in which icon.sys resides.
	strcpy((char*)icon_sys.del, "scummvm.icn");

	// Write icon.sys to the memory card (Note that this filename is fixed)
	mc_fd = fioOpen(fileName ,O_WRONLY | O_CREAT);
	if(mc_fd < 0) return -2;

	fioWrite(mc_fd, &icon_sys, sizeof(icon_sys));
	fioClose(mc_fd);
	//printf("** Create root dir %s written sucessfully.\n", fileName);

	// Write icon file to the memory card.
	// Note: The icon file was created with my bmp2icon tool, available for download at
	//       http://www.ps2dev.org


	strcpy(name, pathName);
	strcat(name, "/scummvm.icn");

	icon_fd = fioOpen(name,O_WRONLY | O_CREAT);
	if(icon_fd < 0) return -6;

	fioWrite(icon_fd,icon_buffer,icon_size);
	fioClose(icon_fd);
	//printf("** Icon written sucessfully.\n");

	return 0;

}

void ps2_get_mcard_dir(int port) {

    extern const char *gGameNameShort;
    int ret;

    char ps2_path[128];
    strcpy(ps2_path, "/SCUMMVM-");
    strcat(ps2_path, gGameNameShort);
    strcat(ps2_path, "/*");

//    clearEntries();
//    _entry_count = 0;

    //prevent freeze
    pause_sound();

    ret = mcGetDir(port, 0, ps2_path, 0, ARRAY_ENTRIES - 10, mcDir);
    if (ret < 0) {
    } else 
    if (ret == 0) //some change occured
      mcSync(0, NULL, &_entry_count);
    //otherwise (3) no change occured - we must not mcSync or else we get 0 entryCount
    printf("**ps2_get_mcard_dir path=%s, mcGetDir ret=%i  entryCount=%i \n",ps2_path, ret, _entry_count);

}

