#include "fcache.h"

#define DPRINTF	 //
//#define DPRINTF printf
/*********************************************************************
 private methods
*********************************************************************/

int strEqual(const char *s1, const char* s2) {
    char u1, u2;
    for (;;) {
		u1 = *s1++;
		u2 = *s2++;
		if (u1 >64 && u1 < 91)  u1+=32;
		if (u2 >64 && u2 < 91)  u2+=32;
        
		if (u1 != u2) {
			return -1;
		}
		if (u1 == '\0') {
		    return 0;
		}
    }
}

FileCache::FileCache(unsigned int maxSize) {
	//FileBlock* fb;

	//20MB limit
	if (maxSize > 20*1024*1024) {
		maxSize = 20*1024*1024;
	}
	DPRINTF("cache: ######!!!!!!###### constructor!!!!!!!!!!!!!!! \n");	
	DPRINTF("creating FileCache size=%d \n", maxSize);
	totalSize = maxSize;
	blockCount = totalSize / MAX_DATA_BLOCK;
	
	//create dummy file block
	//fb = createFileBlock(NULL, NULL);
	fbHead = NULL;
	statCount = 0;
	idCount = 0;
	readCount = 0;
	hitCount = 0;
	lastId[0] = lastId[1] = lastId[2] = lastId[3] = 0;
}

FileCache::~FileCache() {
	FileBlock* fb;
	FileBlock* nextFb;
	DPRINTF("cache: ######!!!!!!###### destructor!!!!!!!!!!!!!!! \n");	
	//iterate through  file blocks and free them
	fb = (FileBlock*)fbHead;
	while (fb != NULL) {
		nextFb = (FileBlock*) fb->next;
		freeFileBlock(fb);
		fb = nextFb;
	}
//	instance = NULL;
}

FileBlock* FileCache::createFileBlock(FILE* file, const char* name, const char* dir) {
	FileBlock* fb;
	int len;
	
	DPRINTF("cache: create file block name=%s dir=%s \n", name, dir);
	fb = (FileBlock*) malloc (sizeof(FileBlock));
	if (!fb) {
		DPRINTF("cache: create file block filed\n");
		return NULL;
	}
	fb->fh = file;
	fb->prev = NULL;
	fb->next = NULL;
	fb->dbHead = NULL;
	fb->hitCount = 0;
	fb->dbCount = 0;
	fb->seekPos = 0;
	if (name != NULL) {
		len = strlen(name);
		fb->fileName = (char*) malloc (len + 1);
		if (fb->fileName) {
			strcpy(fb->fileName, name);
			fb->fileName[len] = 0;
			DPRINTF("cache:  filename copied %s \n", fb->fileName);
		}
	} else {
		fb->fileName = NULL;
	}

	if (dir != NULL) {
		len = strlen(dir);
		fb->directory = (char*) malloc (len + 1);
		if (fb->directory) {
			strcpy(fb->directory, dir);
			fb->directory[len] = 0;
			DPRINTF("cache:  directory copied %s \n", fb->directory);
		}
	} else {
		fb->directory = NULL;
	}

	return fb;
}

// link behind
void FileCache::linkFileBlock(FileBlock* newBlock, FileBlock* currentBlock) {
	FileBlock* nextFb;
	if (newBlock == NULL) {
		return;
	}

	if (currentBlock == NULL) {
		fbHead = newBlock;
	} else {
		nextFb = (FileBlock*) currentBlock->next;
		currentBlock->next  = newBlock;
		newBlock->prev = currentBlock;
		newBlock->next = nextFb;
		if (nextFb) {
			nextFb->prev = newBlock;
		}
	}
}

void FileCache::freeFileBlock(FileBlock* fb) {
	FileBlock* fbPrev, *fbNext;
	DataBlock* db, *dbNext;
	if (fb == NULL) {
		return;
	}
	
	fbPrev = (FileBlock*) fb->prev;
	fbNext = (FileBlock*) fb->next;
	
	//unlink current file block from the chain
	if (fbPrev) {
		fbPrev->next = fbNext;
	} else {
		fbHead = fbNext;
	}
	if (fbNext) {
		fbNext->prev = fbPrev;
	}
	
	//iterate through data blocks and free them
	db = (DataBlock*)fb->dbHead;
	while (db) {
		dbNext = (DataBlock*)db->next;
		freeDataBlock(db);
		db = dbNext;
	}
	
	if (fb->fileName) {
		free(fb->fileName);
	}
	//finally free the file block
	free(fb);
}

DataBlock* FileCache::createDataBlock(FileBlock* fb) {
	DataBlock* db;

	DPRINTF("Create data block \n");	
	//we ran out of allowed space
	if (blockCount <= 0) {
		return NULL;
	}
	
	db = (DataBlock*) malloc (sizeof(DataBlock));
	if (!db) {
		DPRINTF("Create data block failed!1\n");	
		return NULL;
	}
	db->data = (unsigned char*) malloc(MAX_DATA_BLOCK);
	if (!db->data) {
		free(db);
		DPRINTF("Create data block failed!2\n");	
		return NULL;
	}
	//we have succesfully allocated data block and the data buffer
	blockCount --;
	db->prev = NULL;
	db->next = NULL;
	db->offset = 0;
	db->size = 0;
	db->complete = 0;
	db->hitCount = 0;
	db->fileBlock = fb;
	idCount++;
	db->id = idCount;
	DPRINTF("Create data block done. id=%d \n", db->id);	
	return db;
}

void FileCache::freeDataBlock(DataBlock* db) {
	if (db == NULL) {
		return;
	}
	
	if (db->data) {
		free(db->data);
	}
	free(db);
}

int FileCache::isLastId(unsigned int id) {
	if (	id == lastId[3] ||
		id == lastId[2] ||
		id == lastId[1] ||
		id == lastId[0] ) {
			return 1;
	}
	return 0;
}

void FileCache::setLastId(unsigned int id) {
	lastId[0] = lastId[1];
	lastId[1] = lastId[2];
	lastId[2] = lastId[3];
	lastId[3] = id;
}


DataBlock*  FileCache::getDataBlock(FileBlock* destFb, DataBlock* curDb) {
	FileBlock* fb, *parentFb;
	DataBlock* db = NULL;
	DataBlock* minDb = NULL;
	DataBlock* prevDb, *nextDb;
	int minHits  = 1000000;
	int hits;

	//there is enough space
	if (blockCount > 0) {
		db = createDataBlock(destFb);
	}
	//db created succesfully 
	if (db != NULL) {
		return db;
	}

	DPRINTF("!! no more free data !! - get used block \n");	
	//no more data block allowed or not enough space
	//pick the worst (not used) data block
	
	//iterate through all files in the cache
	fb = (FileBlock*)fbHead;
	while (fb != NULL) {
		//decrease hitCount to ensure that hit ratios change in time
		//if file is not accessed frequently then the hitcount stored in the fileBlock
		//decreases in time and data blocks are removed from this fileBlock.
//		fb->hitCount--;
		
		//iterate through data blocks of the file
		db = (DataBlock*) fb->dbHead;
		while (db != NULL) {
			//the same principle as above - searching lowers the hitCount
			//hitting the cache raises the hitCount
			db->hitCount--; 
			hits = fb->hitCount + db->hitCount;
			if (hits < minHits && !isLastId(db->id)) {
				minHits = hits;
				minDb = db;
			}
			db = (DataBlock*) db->next;
		}
		fb = (FileBlock*) fb->next;
	}
	if (minDb == curDb) {
//		printf("cache:  newBlock and curBlock match! \n");
	} else {
	
		//now unlink the db from chain
		prevDb = minDb->prev;
		nextDb = minDb->next;
/*
		int a1, a2, a3;
		a2 = a3 = 0;
		a1 = minDb->id;

		if (nextDb) a3 = nextDb->id;
		if (prevDb) a2 = prevDb->id;
//		printf("cache unlink: cur=%d  (prev=%d  next=%d) file=%s \n", a1, a2, a3, minDb->fileBlock->fileName);
*/
		if (prevDb) {
			prevDb->next = nextDb;
		} else { //prevoius is NULL - so it is the head of the chain - update fileBlock
			parentFb = minDb->fileBlock;
			if (parentFb) {
				parentFb->dbHead = nextDb;
			} else {
				//printf("#####unlink panick! \n");
			}	
		
		}
		if (nextDb) {
			nextDb->prev = prevDb;
		}
	
		minDb->prev = NULL;
		minDb->next = NULL;
		setLastId(minDb->id);
	}
	//clear the data block
	minDb->hitCount = 0;
	minDb->offset = 0;
	minDb->size = 0;
	minDb->complete = 0;
	minDb->fileBlock = destFb;

	return minDb;
}


FileBlock* FileCache::getFileBlockByHandle(FILE* fh) {
	FileBlock* fb;
	
	fb = (FileBlock*)fbHead;
	while (fb != NULL) {
		if (fb->fh == fh) {
			return fb;
		}
		fb = (FileBlock*) fb->next;
	}
	return NULL;
}

FileBlock* FileCache::getFileBlockByName(const char* name, const char* dir, char allowAnyHandle) {
	FileBlock* fb;

	DPRINTF("cache: getFileBlockByName name=%s dir=%s fbHead=%p\n", name, dir, fbHead);	
	if (name == NULL || dir == NULL) {
		return NULL;
	}
	
	fb = (FileBlock*)fbHead;
	if (fb == NULL) {
		DPRINTF("cache: fbHead NULL! \n");
	} else {
		DPRINTF("cache: fbHead Not NULL! \n");
		DPRINTF("cache: fbHead name=%s dir=%s \n", fb->fileName, fb->directory);
	}
	while (fb != NULL) {
		DPRINTF(" cache: comparing name=%s dir=%s \n", fb->fileName, fb->directory);
		if (	strEqual(fb->fileName, name) == 0 && 
			strEqual(fb->directory, dir) == 0) {
				if (fb->fh == NULL || allowAnyHandle) {
					DPRINTF(" cache: compare  success! \n");
					return fb;
				}
		}
		fb = (FileBlock*) fb->next;
	}
	DPRINTF(" cache: compare fail! \n");
	return NULL;
}

void FileCache::linkDataBlockPrior(DataBlock* newBlock, DataBlock* currentBlock) {
	FileBlock* parentFb;
	DataBlock* prevDb;
	int a1, a2, a3;

	a2 = a3 = 0;
	a1 = currentBlock->id;

	if (currentBlock->next) a3 = currentBlock->next->id;
	if (currentBlock->prev) a2 = currentBlock->prev->id;

	//printf("cache linkPrior: new=%d  cur=%d  (prev=%d  next=%d) \n", newBlock->id, a1, a2, a3);
	
	prevDb = (DataBlock*) currentBlock->prev;
	if (prevDb) {
		prevDb->next = newBlock;
	} else {
		parentFb = currentBlock->fileBlock;
		parentFb->dbHead = newBlock;
	}
	
	newBlock->prev = prevDb;
	newBlock->next = currentBlock;
	currentBlock->prev = newBlock;
}

void FileCache::linkDataBlockBehind(DataBlock* newBlock, DataBlock* currentBlock) {
	DataBlock* nextDb;
	int a1, a2, a3;

	a2 = a3 = 0;
	a1 = currentBlock->id;

	if (currentBlock->next) a3 = currentBlock->next->id;
	if (currentBlock->prev) a2 = currentBlock->prev->id;

	//printf("cache linkBehind: new=%d  cur=%d  (prev=%d  next=%d) \n", newBlock->id, a1, a2, a3);

	nextDb = (DataBlock*) currentBlock->next;
	if (nextDb) {
		nextDb->prev = newBlock;
	} 
	newBlock->next = nextDb;
	newBlock->prev = currentBlock;
	currentBlock->next = newBlock;
}

void FileCache::dumpStat() {
	int dbCount;
	int ratio;
	char* dummy;
	FileBlock* fb;
	DataBlock* db;
	int b2, b3;
	
	dbCount = 0;
	printf("Cache statistics \n");
	printf("-----------------\n");
	fb = (FileBlock*) fbHead;
	while (fb != NULL) {
		printf("FB: name=%s dir=%s size=%d pos=%d\n", fb->fileName, fb->directory, fb->fileSize, fb->seekPos);
		db = (DataBlock*) fb->dbHead;
		while (db != NULL) {
			dbCount++;
			b2= b3= 0;
			if (db->prev != NULL) b2 = db->prev->id;
			if (db->next != NULL) b3 = db->next->id;
			printf("  DB: id=%d(%d : %d) hc=%d offset=%d end=%d\n", db->id, b2, b3, db->hitCount, db->offset, db->offset + db->size);
			db = (DataBlock*)db->next;
		}
		fb = (FileBlock*) fb->next;
	}
	dummy= (char*) malloc(8);
	ratio = (hitCount * 100) / readCount;
	printf("readCount=%d hitCount=%d ratio=%d.%d mallocPos=%p dbCount=%d\n", readCount, hitCount, ratio/100, ratio % 100, dummy, dbCount);
	if (dummy) {
		free(dummy);
	}

}

int FileCache::getDbCount() {
	int dbCount;
	FileBlock* fb;
	DataBlock* db;
	
	dbCount = 0;
	fb = (FileBlock*) fbHead;
	while (fb != NULL) {
		db = (DataBlock*) fb->dbHead;
		while (db != NULL) {
			dbCount++;
			db = (DataBlock*)db->next;
		}
		fb = (FileBlock*) fb->next;
	}
	return dbCount;
}


/***************************************************************************
  PUBLIC SECTION
****************************************************************************/

char* FileCache::getFileDir(const char* name) {
	FileBlock* fb;

	DPRINTF("cache: getFileDir name=%s fbHead=%p\n", name, fbHead);	
	if (name == NULL ) {
		return NULL;
	}
	
	fb = (FileBlock*)fbHead;

	if (fb == NULL) {
		DPRINTF("cache: fbHead NULL! \n");
	} 
	while (fb != NULL) {
		if (strEqual(fb->fileName, name) == 0 ) {
			DPRINTF(" cache: compare  success! \n");
			return fb->directory;
		}
		fb = (FileBlock*) fb->next;
	}
	DPRINTF(" cache: compare fail! \n");
	return NULL;
	
}
unsigned int FileCache::getFileSize(const char* name, const char* dir) {
	FileBlock* fb;

	//for Size we can use either null or not null handles
	fb = getFileBlockByName(name, dir, 1);
	if (fb == NULL) {
		DPRINTF("cache: getFileSize failed name=%s dir=%s \n",name, dir);
		return 0;
	} else {
		DPRINTF("cache: getFileSize ok! name=%s dir=%s size=%d\n",name, dir, fb->fileSize);
		return fb->fileSize;
	}
}

unsigned int FileCache::pos(FILE* file) {
	FileBlock* fb;

	fb = getFileBlockByHandle(file);
	if (fb == NULL) {
		DPRINTF("cache: pos failed!\n");	
		return 0;
	}
	DPRINTF("cache: pos called ret=%d\n", fb->seekPos);	
	return fb->seekPos;

}
unsigned int FileCache::getSize(FILE* file) {
	FileBlock* fb;

	fb = getFileBlockByHandle(file);
	if (fb == NULL) {
		DPRINTF("cache: size failed!\n");	
		return 0;
	}
	DPRINTF("cache: size called ret=%d\n", fb->fileSize);	
	return fb->fileSize;
}

int FileCache::seek(FILE* file, int offs, int whence) {
	FileBlock* fb;

	DPRINTF("cache: seek offs=%d whence=%d\n", offs, whence);	
	fb = getFileBlockByHandle(file);
	if (fb == NULL) {
		DPRINTF("cache: setSeekPos failed!\n");	
		return -1;
	}
	switch(whence) {
		case SEEK_SET:
			fb->seekPos = offs;
			break;
		case SEEK_CUR:
			fb->seekPos += offs;
			break;
		case SEEK_END:
			fb->seekPos  = fb->fileSize + offs;
			break;
		default:
			return -1;
	}
	if (fb->seekPos < 0) {
		fb->seekPos = 0;
	}
	if (fb->seekPos > fb->fileSize) {
		fb->seekPos = fb->fileSize;
	}
	return fb->seekPos;
}


int FileCache::open(FILE* file, const char* name, const char* dir, unsigned int size) {
	FileBlock* fb;

	DPRINTF("cache: open name=%s dir=%s \n", name, dir);	
	//allow only NULL handles
	fb = getFileBlockByName(name, dir, 0);
	if (fb == NULL) {
		DPRINTF("cache: open  - create FileBlock\n");	
		fb = createFileBlock(file, name, dir);
		linkFileBlock(fb, (FileBlock*)fbHead);
	} else {
		DPRINTF("cache: open  - already cached\n");	
		if (fb->fh == NULL) {
			fb->fh = file;
			fb->seekPos = 0;
		} else {
			//either the file was not closed or is opened multiple times.
			fb = createFileBlock(file, name, dir);
			linkFileBlock(fb, (FileBlock*)fbHead);
		}
	}
	if (fb == NULL) {
		return -1;
	}
	fb->fileSize = size;

	return 1;
}

int FileCache::close(FILE* file) {
	FileBlock* fb;
	
	DPRINTF("cache: close \n");	
	fb = getFileBlockByHandle(file);
	if (fb == NULL) {
		return -1;
	}
	fb->fh = NULL;
	return 1;
}


int FileCache::read(FILE* file, void* buffer, unsigned int size) {
	FileBlock* fb;
	DataBlock* db, *newDataBlock, *oldDb;
	unsigned int readSize;
	unsigned char* ptr;
	int len;
	char eof, readFlag;
	unsigned int dataStart, dataEnd, dataSize, dataOffs, oldSeek;
//	unsigned int a1,a2,a3, b1,b2,b3;
//	int dbCount1, dbCount2;

/*	
	statCount++;
	if (statCount > 1000) {
		statCount = 0;
		dumpStat();
	}
*/
	readCount++;
	readFlag = 0;

	fb = getFileBlockByHandle(file);
	//panick ? read file that was not opened yet?
	if (fb == NULL) {
		printf("cache:read - file not found \n");
		return -1;
	}
	DPRINTF("cache:read name=%s size=%d seekPos=%d \n", fb->fileName, size, fb->seekPos);	
	
	ptr = (unsigned char*) buffer;
	eof = 0;
	//scan data blocks wheter they contain requested data
	readSize = 0;
	db = (DataBlock*) fb->dbHead;

	//check wether at least one (dbHead) data block exist
	if (db == NULL) {
		//dbCount1 = getDbCount();
		db = getDataBlock(fb, NULL);
		DPRINTF("cache: initial read  db=%p file=%s \n", db->id, fb->fileName);
		//note: we are aligning data block - so the content of data blocks is not overlaped (save space)
		db->offset = (fb->seekPos / MAX_DATA_BLOCK) * MAX_DATA_BLOCK;
		fseek(fb->fh, db->offset, SEEK_SET);
		len = fread(db->data + db->size, 1, MAX_DATA_BLOCK - db->size, fb->fh);
		db->size = len;
		db->complete = 1;
		db->hitCount = 0;
		fb->dbHead = db;
		readFlag = 1;
/*
		dbCount2 = getDbCount();
		if (dbCount1 > dbCount2) {
			statCount = 0;
			dumpStat();
			printf("db lost! prev=%d now=%d \n", dbCount1, dbCount2);
		}
		DPRINTF("cache: initial read done. sizeRead=%d db->size=%d\n", len, db->size);
*/
	}
	
	while (readSize < size && !eof) {
		dataStart = db->offset;
		dataEnd = dataStart + MAX_DATA_BLOCK;
		if (dataEnd >= fb->fileSize) {
			DPRINTF("cache: eof detected! fileSize=%d dataEnd=%d \n",fb->fileSize, dataEnd);
			eof = 1;
		}
		//current data block holds the data we need
		if (dataStart <= fb->seekPos && fb->seekPos < dataEnd ) {
			//check wether the data block is fully loaded - if not then load it
			if (!db->complete) {
				DPRINTF("cache:read name=%s size=%d pos=%d \n", fb->fileName, size, fb->seekPos);	
				DPRINTF("--->cache read - physical read at  pos=%d db=%d hwpos=%d\n", db->offset + db->size, db->id, db->offset);
				oldSeek = fb->seekPos;
				fseek(fb->fh, db->offset + db->size, SEEK_SET);
				len = fread(db->data + db->size, 1, MAX_DATA_BLOCK - db->size, fb->fh);
				fseek(fb->fh, oldSeek, SEEK_SET);
				db->size += len;
				db->complete = 1;
				readFlag = 1;
			}
			dataEnd = dataStart + db->size;
			dataOffs = fb->seekPos - dataStart;
			dataSize = dataEnd - fb->seekPos;
			if (dataSize > (size-readSize)) {
				dataSize = (size-readSize);
			}
			memcpy(ptr + readSize, db->data + dataOffs, dataSize);
			DPRINTF("   cache: memcpy -> dst=%p dst offs=%d  src offs=%d  size=%d data=%02X\n", ptr, readSize, dataOffs, dataSize, *(ptr + readSize));
			readSize += dataSize;
			fb->seekPos +=dataSize;
			db->hitCount++;
			oldDb = db;
			db = db->next;
		}else
		//current data block holds data beyond the current file seek position: data is not in the cache
		//note: data blocks of the file are ordered (the chain) by offset.
		if (dataStart > fb->seekPos) {
			DPRINTF("cache:  current offset larger! \n");
//			dbCount1 = getDbCount();
			newDataBlock = getDataBlock(fb, db);
			//note: we are aligning data block - so the content of data blocks is not overlaped (save space)
			newDataBlock->offset = (fb->seekPos / MAX_DATA_BLOCK) * MAX_DATA_BLOCK;
			//link data block in the current file chain
			//note: the current db might have been just recycled
			//      so we do not have to update the chain
/*
			a1 = a2 = a3 = 0;
			b1 = b2 = b3 = 0;
			if (newDataBlock != NULL) {
				a1= newDataBlock->id;
				if (newDataBlock->prev != NULL) a2 = newDataBlock->prev->id;
				if (newDataBlock->next != NULL) a3 = newDataBlock->next->id;
			}
			if (db != NULL) {
				b1= db->id;
				if (db->prev != NULL) b2 = db->prev->id;
				if (db->next != NULL) b3 = db->next->id;
			}
			DPRINTF("newDB=%d  (p->%d n->%d)   oldDB=%d (p->%d n->%d)\n", a1, a2, a3, b1, b2, b3);
*/
			if (newDataBlock != db) {
				linkDataBlockPrior(newDataBlock, db);
/*
				dbCount2 = getDbCount();
				if (dbCount1 > dbCount2) {
					statCount = 0;
					dumpStat();
					printf("db lost! prev=%d now=%d \n", dbCount1, dbCount2);
				}
*/
			}

			db = newDataBlock; //in the next cycle this db will be processed (and loaded)
			eof = 0; //clear eof to prevent early exit
		} else{
/*
			if (db->next == NULL) {
			        DPRINTF("no more data blocks found. \n");
			} else {
				DPRINTF("probing next db id=%d db offset=%d\n", db->next->id, db->next->offset);
			}
*/
			oldDb = db;
			db = db->next;
		}

		DPRINTF(" cache read next iteration: db=%p readSize=%d size=%d eof=%d \n", db, readSize, size, eof);
		if (db == NULL && readSize < size && !eof) {
//			dbCount1 = getDbCount();
			newDataBlock = getDataBlock(fb, oldDb);
			DPRINTF("cache read -  end of  data blocks newBlock=%p \n", newDataBlock->id);
			//note: we are aligning data block - so the content of data blocks is not overlaped (save space)
			newDataBlock->offset = (fb->seekPos / MAX_DATA_BLOCK) * MAX_DATA_BLOCK;
			//link data block in the current file chain
/*
			a1 = a2 = a3 = 0;
			b1 = b2 = b3 = 0;
			if (newDataBlock != NULL) {
				a1= newDataBlock->id;
				if (newDataBlock->prev != NULL) a2 = newDataBlock->prev->id;
				if (newDataBlock->next != NULL) a3 = newDataBlock->next->id;
			}
			if (oldDb != NULL) {
				b1= oldDb->id;
				if (oldDb->prev != NULL) b2 = oldDb->prev->id;
				if (oldDb->next != NULL) b3 = oldDb->next->id;
			}
			DPRINTF("newDB=%d  (p->%d n->%d)   oldDB=%d (p->%d n->%d)\n", a1, a2, a3, b1, b2, b3);
*/
			if (newDataBlock != oldDb) {
				linkDataBlockBehind(newDataBlock, oldDb);
/*
				dbCount2 = getDbCount();
				if (dbCount1 > dbCount2) {
					statCount = 0;
					dumpStat();
					printf("db lost! prev=%d now=%d \n", dbCount1, dbCount2);
				}
*/
			}
			db = newDataBlock; //in the next cycle this db will be processed (and loaded)
		}
	}
//	printf("read ok. sizeRequested=%d sizeRead=%d \n",size, readSize);
	if (!readFlag) {
		hitCount++;
	}
//	fb->hitCount++;
	return readSize;
}
