#ifndef _FILE_CACHE_H_
#define _FILE_CACHE_H_

#include "stdafx.h"


/* the size of allocated space (in bytes) per data block */
#define MAX_DATA_BLOCK 65536

typedef struct _file_block {
	void*		prev;		//previous file block
	void*		next;		//next file block
	void*		dbHead;		//first data block
	char*		fileName;	
	char*		directory;
	FILE*		fh;		//file handle
	unsigned int	hitCount;	//number of file hits
	unsigned int	dbCount;	//number of attached data blocks (to qickly compute allocated space)
	unsigned int	seekPos;	//seek position in file
	unsigned int	fileSize;	//total size of the file
} FileBlock;

typedef struct _data_block {
	struct _data_block*		prev;		//previous data block
	struct _data_block*		next;		//next data block
	FileBlock*      fileBlock;	//parent file block 
	unsigned char*	data;		//data memory block of size=MAX_DATA_BLOCK 
	unsigned int	offset;		//file offset of the data block first byte
	int 		size;		//size of the filled data block (should be MAX_DATA_BLOCK when data are completely filled)
	int 		complete;	//0-data partially loade 1-data fully loaded
	unsigned int 	hitCount;	//number of data block hits
	unsigned int    id;
} DataBlock;


class FileCache  {
private:
//	static FileCache* instance;
	
	unsigned int 	totalSize;	//max size of the cache in bytes
	int		blockCount;	//data block count
	void*		fbHead;		//first file block
	unsigned int	statCount;
	unsigned int	idCount;
	unsigned int    readCount;
	unsigned int    hitCount;
	unsigned int	lastId[4];

	void 		dumpStat();	//dump statistic info
	int 		getDbCount();
	FileBlock*	createFileBlock(FILE* file, const char* name, const char* dir);
	DataBlock*	createDataBlock(FileBlock* fb);
	void 		freeFileBlock(FileBlock* fb);
	void 		freeDataBlock(DataBlock* db);
	int 		isLastId(unsigned int id);
	void 		setLastId(unsigned int id);
	
	
	FileBlock*	getFileBlockByHandle(FILE* fh);
	FileBlock*	getFileBlockByName(const char* name, const char* dir, char nullHandle);
	DataBlock*	getDataBlock(FileBlock* fb, DataBlock* curBlock); 
	
	void		linkDataBlockPrior(DataBlock* newBlock, DataBlock* currentBlock);
	void 		linkDataBlockBehind(DataBlock* newBlock, DataBlock* currentBlock);
	void		linkFileBlock(FileBlock* newBlock, FileBlock* currentBlock);

public:
	FileCache(unsigned int maxSize);
	~FileCache();
	
//	static FileCache* getInstance(unsigned int maxSize);

	int open(FILE* file, const char* name, const char* dir, unsigned int size);
	int close(FILE* file);
	int read(FILE* file, void* buffer, unsigned int size);
	//int write(FILE* file, void* buffer, unsigned int size);
//	int setSeekPos(FILE* file, unsigned int pos);
	int seek(FILE* file, int offs, int whence);
	unsigned int pos(FILE* file);
	unsigned int getSize(FILE* file);
	char* getFileDir(const char* name);
	unsigned int getFileSize(const char* name, const char* dir);

	//const char* getFileName(const char name);
};

#endif /* _FILE_CACHE_H_ */
